from jax.numpy import sign as sign_function
from jax import custom_vjp


def sign(function):
    wrapper = custom_vjp(function)

    def forward(x):
        return function(x), x

    def backward(x, gradient):
        return (gradient * sign_function(x),)

    wrapper.defvjp(forward, backward)
    return wrapper



from functools import partial
from jax.tree_util import tree_map
from jax import grad
from .loss import loss
from .gd import gd


def fit(learning_rate, model, x, y):
    gradients = grad(loss)(model, x, y)
    return tree_map(partial(gd, learning_rate), model, gradients)



from . import vjp
from .abs import abs
from .apply import apply
from .bias import bias
from .binary import binary
from .bind import bind
from .box import Box
from .compose import compose
from .concat import concat
from .create import create
from .explode import explode
from .fit import fit
from .gd import gd
from .identity import identity
from .implode import implode
from .loss import loss
from .model import Model
from .mse import mse
from .param import param
from .phi import PHI
from .relu import relu
from .relu1 import relu1
from .relun import relun
from .split import split
from .stack import stack
from .unstack import unstack
from .vector import vector
from .weight import weight

Neuron = bind("Neuron", create, apply)
Binary = compose("Binary", Neuron, binary)
Vector = compose("Vector", Neuron, vector)

__all__ = [
    "abs",
    "apply",
    "bias",
    "binary",
    "Binary",
    "bind",
    "Box",
    "compose",
    "concat",
    "create",
    "explode",
    "fit",
    "gd",
    "identity",
    "implode",
    "loss",
    "Model",
    "mse",
    "Neuron",
    "param",
    "PHI",
    "relu",
    "relu1",
    "relun",
    "split",
    "stack",
    "unstack",
    "vector",
    "Vector",
    "vjp",
    "weight",
]
